from .lexer import Lexer, LexError
from .token import Token
from .parser import Parser, ParseError
from .makefile import Makefile

import sys
import copy
import pathlib
import colored
import json


def main(argc: int = len(sys.argv), argv: list[str] = sys.argv) -> int:
    if argc < 2:
        print("muild: %s: no input directory" %
              (colored.stylize("error", colored.fg(1))))

        return 1

    config_fname = "build.txt"

    simple = pathlib.Path(argv[1])
    direct = simple.absolute().resolve()
    config = direct / config_fname
    cwd = pathlib.Path(".").absolute().resolve()
    home_path = pathlib.Path.home()
    lib_path = home_path / ".local" / "lib" / "muild"

    if not lib_path.exists():
        print("muild: %s: invalid installation" %
              (colored.stylize("error", colored.fg(1))))

        return 1

    if not direct.is_dir():
        print("muild: %s: '%s' is not a directory" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(simple, colored.fg(15))))

        return 1

    if not config.is_file():
        print("muild: %s: '%s' is not a file" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(simple / config_fname, colored.fg(15))))

        return 1

    with config.open("r") as config_fp:
        config_fc: str = config_fp.read()

    if not config_fc:
        print("muild: %s: '%s' is empty" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(simple / config_fname, colored.fg(15))))

        return 1

    lexer = Lexer(config_fc)
    tokens: list[Token] = []

    try:
        for token in lexer:
            tokens.append(token)

    except LexError as lexer_error:
        print("muild: %s: %s" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(lexer_error, colored.fg(15))))

        return 1

    parser = Parser(tokens)

    try:
        project = parser.parse()

    except ParseError as parser_error:
        print("muild: %s: %s" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(parser_error, colored.fg(15))))

        return 1

    for submod in project["submods"]:
        project = configure_submod(project, submod, direct)

    for setup in project["setups"]:
        setup_path: pathlib.Path = lib_path / (setup + ".json")

        if not setup_path.exists():
            print("muild: %s: '%s' is not a file" %
                  (colored.stylize("error", colored.fg(1)), colored.stylize(setup_path, colored.fg(15))))

            return 1

        with setup_path.open("r") as setup_fp:
            setup_fc: str = setup_fp.read()

        try:
            setup_c: dict = json.loads(setup_fc)

        except json.decoder.JSONDecodeError as json_error:
            print("muild: %s: %s" %
                  (colored.stylize("error", colored.fg(1)), colored.stylize(json_error, colored.fg(15))))

            return 1

        for name in setup_c:
            project["variables"][name.upper()] = setup_c[name]

    project["variables"]["CXXFLAGS"] += " -I%s -I%s" % (direct, cwd)

    with Makefile("Makefile", comment="Auto-generated by muild") as makefile:
        makefile.file.write(".DEFAULT_GOAL := all\n")

        for variable in project["variables"]:
            makefile.add_macro(variable, project["variables"][variable])

        outs: list[pathlib.Path] = []
        targets: list[str] = []

        for target in project["targets"]:
            _src: list[pathlib.Path] = []
            _body: list[str] = []
            _deps: list[str] = []
            _links: list[str] = []
            _outs: list[str] = []

            for dep in project["targets"][target]["deps"]:
                if dep.startswith(":"):
                    _deps.append(
                        "lib" + project["name"] + "-" + dep[1:] + ".a")

                else:
                    _links.append("-l" + dep)

            for source in project["targets"][target]["src"]:
                _src.append(direct / source)

            for source in _src:
                out_name = cwd / (source.name + ".o")
                outs.append(out_name)
                _outs.append(str(out_name))

                _body.append("$(CXX) $(CXXFLAGS) -c %s -o %s" %
                             (source, out_name))

            if project["targets"][target]["type"] == "executable":
                _name = project["name"] + "-" + target

                _body.append("$(CXX) $(CXXFLAGS) $(LDFLAGS) %s -o %s %s %s" %
                             (" ".join(_outs), _name, " ".join(_links), " ".join(_deps)))

            elif project["targets"][target]["type"] == "static":
                _name = "lib" + project["name"] + "-" + target + ".a"

                _body.append("$(AR) rcs %s %s" %
                             (_name, " ".join(_outs)))

            elif project["targets"][target]["type"] == "shared":
                # TODO: Implement shared library support
                print("muild: %s: shared libraries are not supported yet" %
                      (colored.stylize("error", colored.fg(1))))

                return 1

            else:
                print("muild: %s: invalid target type '%s'" %
                      (colored.stylize("error", colored.fg(1)), colored.stylize(target, colored.fg(15))))

                return 1

            targets.append(_name)
            makefile.add_target(_name, _deps, _body)

        makefile.add_target("all", targets, [])

        clean_body: list[str] = []

        for out in outs:
            clean_body.append(f"$(RM) {out}")

        for target in targets:
            clean_body.append(
                f"$(RM) {pathlib.Path(target).absolute().resolve()}")

        makefile.add_target("clean", [], clean_body)

    return 0


def configure_submod(project: dict, submod: str, direct: pathlib.Path) -> dict:
    project_copy = copy.deepcopy(project)

    submod_path: pathlib.Path = direct / submod

    if not submod_path.is_dir():
        print("muild: %s: '%s' is not a directory" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(submod_path, colored.fg(15))))

        sys.exit(1)

    submod_path = submod_path.absolute().resolve()
    submod_config = submod_path / "build.txt"

    if not submod_config.is_file():
        print("muild: %s: '%s' is not a file" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(submod_config, colored.fg(15))))

        sys.exit(1)

    submod_fc: str = submod_config.read_text()

    try:
        lexer = Lexer(submod_fc)
        tokens: list[Token] = []

        for token in lexer:
            tokens.append(token)

    except LexError as lexer_error:
        print("muild: %s: %s" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(lexer_error, colored.fg(15))))

        sys.exit(1)

    try:
        parser = Parser(tokens)
        submod_project = parser.parse()

    except ParseError as parser_error:
        print("muild: %s: %s" %
              (colored.stylize("error", colored.fg(1)), colored.stylize(parser_error, colored.fg(15))))

        sys.exit(1)

    if submod_project["name"] != None:
        print("muild: %s: cannot use project() in submodule" %
              (colored.stylize("error", colored.fg(1))))

        sys.exit(1)

    for variable in submod_project["variables"]:
        project_copy["variables"][variable] = submod_project["variables"][variable]

    for target in submod_project["targets"]:
        project_copy["targets"][target] = submod_project["targets"][target]

    for submod in submod_project["submods"]:
        submod_project = configure_submod(project_copy, submod, direct)

    return project_copy
