from .position import Position
from .token import Token

import io
import copy
import json5

reserved = {
    "project": "key_project",
    "import": "key_import",
    "library": "key_library",
    "static": "key_static",
    "shared": "key_shared",
    "executable": "key_executable",
    "setup": "key_setup",
}


class LexError(BaseException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class Lexer:
    __source: str
    __position: Position
    __current: str

    def __init__(self, source: str | io.TextIOWrapper) -> None:
        if isinstance(source, str):
            self.__source = source

        elif isinstance(source, io.TextIOWrapper):
            self.__source = source.read()

        self.__position = Position()

        try:
            self.__current = self.__source[int(self.__position)]

        except IndexError as index_error:
            raise LexError("empty source") from index_error

    @property
    def source(self) -> str:
        return copy.copy(self.__source)

    @property
    def position(self) -> Position:
        return copy.copy(self.__position)

    @property
    def current(self) -> str:
        return copy.copy(self.__current)

    def advance(self) -> None:
        if int(self.__position) + 1 > len(self.__source):
            self.__current = None

        self.__position.advance(self.__current)

        try:
            self.__current = self.__source[int(self.__position)]

        except IndexError:
            self.__current = None

    def __iter__(self) -> "Lexer":
        return self

    def __next__(self) -> Token:
        while self.__current is not None and self.__current in " \n\t\r":
            self.advance()

        if self.__current is None:
            raise StopIteration

        elif self.__current in "qwertyuopasdfghjklizxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM_":
            position: Position = self.position
            value: str = ""

            while self.__current is not None and self.__current in "qwertyuopasdfghjklizxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM0123456789_":
                value += self.__current

                self.advance()

            type = "id"

            if value in reserved:
                type = reserved[value]

            return Token(type, value, position)

        elif self.__current in "0123456789":
            position: Position = self.position
            value: str = ""

            while self.__current is not None and self.__current in "0123456789.":
                value += self.__current

                self.advance()

            if "." in value:
                if value.endswith("."):
                    raise LexError(f"invalid float {value}")

                return Token("float", float(value), position)

            return Token("int", int(value), position)

        else:
            match self.__current:
                case "/":
                    position: Position = self.position
                    self.advance()

                    if self.__current is None:
                        raise LexError("unexpected end of file")

                    elif self.__current == "/":
                        while self.__current is not None and self.__current != "\n":
                            self.advance()

                        if self.__current is None:
                            raise LexError("unexpected end of file")

                        else:
                            self.advance()

                            return self.__next__()

                    elif self.__current == "*":
                        self.advance()

                        while self.__current is not None:
                            if self.__current == "*":
                                self.advance()

                                if self.__current is None:
                                    raise LexError("unexpected end of file")

                                elif self.__current == "/":
                                    self.advance()

                                    return self.__next__()

                            else:
                                self.advance()

                        raise LexError("unexpected end of file")

                    else:
                        # Unused token 'div'
                        return Token("div", "/", position)

                case "\"":
                    position: Position = self.position
                    value: str = ""

                    self.advance()

                    while self.__current is not None and self.__current != "\"":
                        value += self.__current

                        self.advance()

                    self.advance()

                    return Token("string", value, position)

                case "{":
                    position: Position = self.position
                    value: str = ""

                    while self.__current is not None and self.__current != "}":
                        value += self.__current

                        self.advance()

                    value += self.__current

                    self.advance()

                    return Token("map", json5.loads(value), position)

                case '(':
                    position: Position = self.position
                    self.advance()

                    return Token("lparen", "(", position)

                case ')':
                    position: Position = self.position
                    self.advance()

                    return Token("rparen", ")", position)

                case ',':
                    position: Position = self.position
                    self.advance()

                    return Token("comma", ",", position)

                case '=':
                    position: Position = self.position
                    self.advance()

                    return Token("assign", "=", position)

                case "[":
                    position: Position = self.position
                    value: str = ""

                    self.advance()

                    while self.__current is not None and self.__current != "]":
                        value += self.__current

                        self.advance()

                    self.advance()

                    return Token("condition", value, position)

                # case ':':
                #     position: Position = self.position
                #     self.advance()

                #     return Token("colon", ":", position)

                case _:
                    raise LexError(
                        f"unsupported character '{self.__current}'")

    def __copy__(self) -> "Lexer":
        lexer: Lexer = Lexer(copy.copy(self.__source))

        lexer.__position = copy.copy(self.__position)
        lexer.__current = copy.copy(self.__current)

        return lexer

    def __deepcopy__(self, memo: dict) -> "Lexer":
        lexer: Lexer = Lexer(copy.deepcopy(self.__source, memo))

        lexer.__position = copy.deepcopy(self.__position, memo)
        lexer.__current = copy.deepcopy(self.__current, memo)

        return lexer

    def __repr__(self) -> str:
        return f"Lexer({repr(self.__source)}, {repr(self.__position)}, {repr(self.__current)})"

    def __str__(self) -> str:
        return self.__repr__()
