import os
import copy
import typing
import io


class Target:
    __name: str
    __deps: list[str]
    __body: list[str]

    def __init__(self, name: str, deps: list[str], body: list[str]) -> None:
        self.__name = name
        self.__deps = deps
        self.__body = body

    @property
    def name(self) -> str:
        return copy.copy(self.__name)

    @property
    def dependencies(self) -> list[str]:
        return copy.deepcopy(self.__deps)

    @property
    def body(self) -> list[str]:
        return copy.deepcopy(self.__body)

    def __repr__(self) -> str:
        return f"Target({self.__name}, {self.__deps}, {self.__body})"

    def __str__(self) -> str:
        return "%s: %s\n\t%s" % (self.__name, ' '.join(self.__deps), '\n\t'.join(self.__body))


class Macro:
    __name: str
    __value: str

    def __init__(self, name: str, value: str) -> None:
        self.__name = name
        self.__value = value

    @property
    def name(self) -> str:
        return copy.copy(self.__name)

    @property
    def value(self) -> str:
        return copy.copy(self.__value)

    def __repr__(self) -> str:
        return f"Macro({self.__name}, {self.__value})"

    def __str__(self) -> str:
        return f"{self.__name}={self.__value}"


class Makefile:
    __name: str
    __file: typing.TextIO
    __targets: list[Target]
    __macros: list[Macro]
    __comment: str

    def __init__(self, name: str = "Makefile", comment: str = "Generated using <class muild.makefile.Makefile>") -> None:
        self.__name = name
        self.__file = None
        self.__targets = []
        self.__macros = []
        self.__comment = comment

    @property
    def name(self) -> str:
        return copy.copy(self.__name)

    @property
    def file(self) -> typing.TextIO:
        """Get ``self.__file``.

        No copy done in this property.
        """
        return self.__file

    @property
    def targets(self) -> list[Target]:
        return copy.deepcopy(self.__targets)

    @property
    def macros(self) -> list[Macro]:
        return copy.deepcopy(self.__macros)

    @property
    def comment(self) -> str:
        return copy.copy(self.__comment)

    def dump(self) -> None:
        self.__file.write(self.dumps())

    def add_target(self, name: str, deps: list[str], body: list[str]) -> None:
        self.__targets.append(Target(name, deps, body))

    def add_macro(self, name: str, value: str) -> None:
        self.__macros.append(Macro(name, value))

    def dumps(self) -> str:
        macros = []
        targets = []

        for macro in self.__macros:
            macros.append(str(macro))

        for target in self.__targets:
            targets.append(str(target))

        return "\n# %s\n\n%s\n\n%s\n" % (self.__comment, "\n".join(macros), "\n\n".join(targets))

    def __enter__(self) -> "Makefile":
        self.__file = io.open(self.__name, "w+")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dump()
        self.__file.close()

    def __repr__(self) -> str:
        return f"Makefile({repr(self.__name)}, {repr(self.__file)})"

    def __str__(self) -> str:
        return self.dumps()
