from .token import Token
from .lexer import Lexer

import copy
import json5


class ParseError(BaseException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class Parser:
    __source: list[Token]
    __position: int
    __current: Token

    def __init__(self, source: list[Token]) -> None:
        self.__source = source
        self.__position = 0

        try:
            self.__current = self.__source[self.__position]

        except IndexError as index_error:
            raise ParseError("empty source") from index_error

    @property
    def source(self) -> list[Token]:
        return copy.deepcopy(self.__source)

    @property
    def position(self) -> int:
        return copy.copy(self.__position)

    @property
    def current(self) -> Token:
        return copy.copy(self.__current)

    def advance(self) -> None:
        if self.__position + 1 > len(self.__source):
            self.__current = None

        self.__position += 1

        try:
            self.__current = self.__source[self.__position]

        except IndexError:
            self.__current = None

    def parse(self) -> dict:
        result: dict = {
            "name": None,
            "targets": {},
            "variables": {},
            "submods": [],
            "setups": [],
        }

        while self.__current is not None:
            if self.__current.type == "key_project":
                self.advance()

                if self.__current.type == "lparen":
                    self.advance()

                    if self.__current.type == "string":
                        result["name"] = self.__current.value

                        self.advance()

                        if self.__current.type == "rparen":
                            self.advance()

                            continue

                        else:
                            raise ParseError("unclosed parenthesis")

                    else:
                        raise ParseError("expected a string")

                else:
                    raise ParseError("unexpected token")

            elif self.__current.type == "key_import":
                self.advance()

                if self.__current.type == "lparen":
                    self.advance()

                    if self.__current.type == "string":
                        if self.__current.value in result["submods"]:
                            raise ParseError("duplicate submodule")

                        result["submods"].append(self.__current.value)

                        self.advance()

                        if self.__current.type == "rparen":
                            self.advance()

                            continue

                        else:
                            raise ParseError("unclosed parenthesis")

                    else:
                        raise ParseError("expected a string")

                else:
                    raise ParseError("unexpected token")

            elif self.__current.type == "key_library":
                self.advance()

                if self.__current.type == "lparen":
                    self.advance()

                    if self.__current.type == "string":
                        name: str = self.__current.value

                        self.advance()

                        if self.__current.type == "comma":
                            self.advance()

                            if self.__current.type == "key_static":
                                type: str = "static"

                            elif self.__current.type == "key_shared":
                                type: str = "shared"

                            else:
                                raise ParseError(
                                    f"invalid library type '{self.__current.value}' (expected static or shared)")

                            self.advance()

                            if self.__current.type == "rparen":
                                self.advance()

                                if self.__current.type == "map":
                                    configuration: dict = self.__current.value

                                    try:
                                        result["targets"][name] = {
                                            "type": type,
                                            "src": configuration["src"],
                                            "deps": configuration["deps"],
                                        }

                                    except KeyError as key_error:
                                        as_str: str = json5.dumps(
                                            self.__current.value, indent=4)

                                        raise ParseError(
                                            f"invalid library configuration '{as_str}' (expected 'src' and 'deps')") from key_error

                                    self.advance()

                                    continue

                                else:
                                    raise ParseError("expected a map")

                            else:
                                raise ParseError("unclosed parenthesis")

                        else:
                            raise ParseError("expected ','")

                    else:
                        raise ParseError("expected a string")

                else:
                    raise ParseError("unexpected token")

            elif self.__current.type == "key_executable":
                self.advance()

                if self.__current.type == "lparen":
                    self.advance()

                    if self.__current.type == "string":
                        name: str = self.__current.value

                        self.advance()

                        if self.__current.type == "rparen":
                            self.advance()

                            if self.__current.type == "map":
                                configuration: dict = self.__current.value

                                try:
                                    result["targets"][name] = {
                                        "type": "executable",
                                        "src": configuration["src"],
                                        "deps": configuration["deps"],
                                    }

                                except KeyError as key_error:
                                    as_str: str = json5.dumps(
                                        self.__current.value, indent=4)

                                    raise ParseError(
                                        f"invalid executable configuration '{as_str}' (expected 'src' and 'deps')") from key_error

                                self.advance()

                                continue

                            else:
                                raise ParseError("expected a map")

                        else:
                            raise ParseError("unclosed parenthesis")

                    else:
                        raise ParseError("expected a string")

                else:
                    raise ParseError("unexpected token")

            elif self.__current.type == "key_setup":
                self.advance()

                if self.__current.type == "lparen":
                    self.advance()

                    if self.__current.type == "string":
                        if self.__current.value in result["setups"]:
                            raise ParseError("duplicate setup")

                        result["setups"].append(self.__current.value)

                        self.advance()

                        if self.__current.type == "rparen":
                            self.advance()

                            continue

                        else:
                            raise ParseError("unclosed parenthesis")

                    else:
                        raise ParseError("expected a string")

                else:
                    raise ParseError("unexpected token")

            elif self.__current.type == "id":
                name: str = self.__current.value
                self.advance()

                if self.__current.type == "assign":
                    self.advance()

                    if self.__current.type == "string":
                        result["variables"][name] = self.__current.value

                        self.advance()

                        continue

                    else:
                        raise ParseError("expected a string")

                else:
                    raise ParseError("unexpected token")

            else:
                raise ParseError(f"unexpected token {self.__current}")

        return result

    def __copy__(self) -> "Parser":
        parser: Parser = Parser(copy.copy(self.__source))

        parser.__position = copy.copy(self.__position)
        parser.__current = copy.copy(self.__current)

        return parser

    def __deepcopy__(self, memo: dict) -> "Parser":
        parser: Parser = Parser(copy.deepcopy(self.__source, memo))

        parser.__position = copy.deepcopy(self.__position, memo)
        parser.__current = copy.deepcopy(self.__current, memo)

        return parser

    def __repr__(self) -> str:
        return f"Parser({repr(self.__source)}, {repr(self.__position)}, {repr(self.__current)})"

    def __str__(self) -> str:
        return self.__repr__()
